/*
  GM Arts program for BJ Devices TB-12
*/

/*===================
  VERSION HISTORY
  ===================

  v0.01
  first version

  v0.02
  using LCD library based on one from extreme electronics
  added code to read pedals and send heel-down CC on pedal #1
  using timer to flash & chase mode LEDs
  bug fixes
  streamlined code

  v1.00
  added pedal heel effect for pedal #2
  added "pedal panic" option
  bug fixes - no known bugs with current config

  v1.01
  arranged config data into byte vars (to simplify saving into EEPROM later)
  added pedal calibration setup
  included config settings validation
  added mutually exclusive effect groups
  added option for persistent effects
  added relay control for banks and effects
  custom message (after patch change) can send 2 x CC messages

  v1.02
  added relays.h and relays.c, added opposite direction fx order index array, using array for bank character A to E
  added tap reset (so that tap must be pressed the preset number of times without intervening presses of other footswitches)
  config: swapped F5 and F9 fx locations
  current patch LED flashes if not in current bank (ie a different bank has been selected for the next patch)

  v1.03
  replaced bool vars & functions with uint8_t (led.c still has a couple of bool function parameters)
  using bit flags instead of FxNow[] and FxPrev[] arrays, also using one FX CC message array for all dual CC message configs
  pedal heel-down FX can now send 2 x CC messages (or 1 x CC and set relay states)
  "Select Bank" displayed in bank select mode
  fixed bug that prevented relay #4 being set

  v2.00
  displays Axe-FX Tuner automatically, reverts to normal display if no tuner msg received for 1/2 sec
  added more pedal history to prevent jitter
  updated midi.c & midi.h files (2/4/16)
  LCD shows arrow on top line when FX are shifted

  v2.10 beta
  can now send custom CC message either before or after CC resets (after a patch change)
  rearranged settings memory locations
  settings will load from EEPROM if correct setting version, else uses defaults
  settings validation restricts relay settings to 2nd "CC message"
  if min pedal calibration setting >= max, resets to full ADC range 0 to 255
  fixed double use of bit 7 of Conf_Pedal_Panic[] config var
  minor updates to lcd_tb.c/h that don't affect this code
  small change to tuner sensitivity (slightly less sensitive near centre than v2.00, but still more sensitive near centre)
  added Axe-FX II tempo backtracking

  v2.10
  Effects F1 to F9 always send on message when pressed if toggles with no off message

  v2.11
  minor code improvements
  using updated lcd_tb.c which uses a different method for custom characters

*/

// ===================
// INCLUDES
// ===================

#include "bjdevlib_tb.h"  // modified for this version to include "adc" instead of "expression" (same for bjdevlib_tb.c)
#include "axefx.h"
#include "lcd_tb.h"       // extreme electronics lcd library adapted for TB-12 with some additions
#include "relays.h"       // GM Arts interface for TB-12 relays

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

// ===================
// CONSTANTS
// ===================

const char *Display_Version = "by GM Arts v2.11";

#define false0  0
#define true1   1

typedef enum ModeType{
  MODE_STARTUP = 0    // Before a patch is selected, bottom row fsw can be used to select a CC message that will be sent after each PC
  ,MODE_PLAY          // PLAY mode for patches and FX - bFxShift var stores FX shift state
  ,MODE_BANK          // fsw are used to select one of 25 banks and return to previous mode
  ,MODE_SETUP         // setup mode using press-buttons
  ,MODE_SETUP_NOC     // setup mode also, when no cursor is required
}ModeType;

#define FX9_INDEX 8   // value in FswToFx[] array for FX9
#define TAP_INDEX 9   // value in FswToFx[] array for Tap

#define NONE     255  // for CC numbers or values, means "no CC message"
#define PREV     254  // for pedal CC values sent after a PC, means "send previous pedal value"
#define RELAYS   128  // when used as a CC number, CC on and off values use Rn_ constants to set relay states

// LED groups can be updated by using one or more of the groups below
#define LEDS_1_TO_5    1
#define LEDS_7_TO_11   2
#define LEDS_6_AND_12  4
#define LEDS_ALL       7

// keypad buttons
#define KEY_RIGHT    12 //INC
#define KEY_LEFT     13 //DEC
#define KEY_UP       14
#define KEY_DOWN     15
#define KEY_LOAD_OK  16
#define KEY_SETUP_NO 17

// Relay values to combine in config data (if you omit a relay, its state does not change)
#define R1_OFF  0b00010000
#define R1_ON   0b00010001
#define R2_OFF  0b00100000
#define R2_ON   0b00100010
#define R3_OFF  0b01000000
#define R3_ON   0b01000100
#define R4_OFF  0b10000000
#define R4_ON   0b10001000
// some useful relay combos
#define REL_ALL_OFF  0b11110000
#define REL_ALL_ON   0b11111111
#define R1_ONLY  0b11110001     // all other relays off
#define R2_ONLY  0b11110010     // all other relays off
#define R3_ONLY  0b11110100     // all other relays off
#define R4_ONLY  0b11111000     // all other relays off

/*
  ===================
  FOOTSWITCH LAYOUT
  ===================

  OVERVIEW
  > 25 banks of 5 patches (125 patches total), any patch can call any MIDI program number, optionally with MIDI banks
  > 10 effects (IAs) including Tap, each can send two CC messages with any off and on value, can be momentary or toggle
  > 2 expression pedals, pedal #1 has heel down effect (eg turns tuner on when volume is at minimum)
  > Backtracking, including effect states and auto bank-change
  > Can select a CC message to be sent after each patch selection
  > And extra "GM Arts smarts" :-)

  Settings below give this footswitch (fsw) layout:

  STARTUP MODE
  Before a patch is sent, can select a custom message with C1 to C6
  (A custom message is previewed here, and if one is selected, it is sent after each patch change)
      [P1]  [P2]  [P3]  [P4]  [P5]  [BANK]
      [C1]  [C2]  [C3]  [C4]  [C5]  [C6]

  PLAY MODE
  After a patch is sent, can select patches and effects (incl Tap):
      [P1]  [P2]  [P3]  [P4]  [P5]  [BACK]
      [F1]  [F2]  [F3}  [F4]  [Tap] [FX shift]

  [FX shift] toggles bottom row FX fsw between above and below
      [P1]  [P2]  [P3]  [P4]  [P5]  [BACK]
      [F5]  [F6]  [F7}  [F8]  [F9]  [FX unshift]

  BANK MODE is selected from STARTUP mode or by holding down [BACK] in PLAY mode
  Bank is selected with group A to E and bank number 1 to 5 (which also returns to previous mode)
      [1]  [2]  [3]  [4]  [5]  [EXIT]
      [A]  [B]  [C]  [D]  [E]  [NOW]
  [EXIT] returns to previous mode with no bank change
  If a patch has been selected, [NOW] selects bank containing that patch and returns to PLAY mode

  ***** PLAY MODE LCD DISPLAY *****

    Bank A5   5--89   // shows bank (for next patch selection) and shifted effects
  Patch A3-2 12-4T   // shows last selected bank & patch and unshifted effects (T is for tap)
*/

// ===================
// CONFIG DEFAULTS (will be replaced by EEPROM settings if corrent settings version)
// ===================

// NOTES:
//  - CC numbers and values below can be set to NONE (value 255)
//  - CC numbers must not be 0 or 32 (these are MIDI MSB and LSB banks reserved for patch selection)
//  - CC numbers for the second message (except pedal CC numbers) can be RELAYS (= 128)
//    and on and off values use Rn_ constants to set relay states

// ****** VERSION ******

uint8_t Conf_Settings_Version[2] = {1, 1};  //<0> settings version (MSB saved in [0]) - may be used by several program versions

// ****** MIDI ******
uint8_t Conf_MIDI_Channel = 0;      //<2> 0-based MIDI channel 0 to 15 for MIDI channels 1 to 16
uint8_t Conf_MIDI_Bank_MSB = 0;     //<3> MIDI MSB PC Bank 0 to 127, or 255 for none
uint8_t Conf_MIDI_Bank_LSB = NONE;  //<4> MIDI LSB PC Bank 0 to 127, or 255 for none

// ****** PROGRAMS ******
// 5 fsw on top row (7 to 11) are preset change buttons
// Here are their 0-based MIDI preset numbers for the 5 patches, each in 25 banks
#define F9ON 0b10000000 // this can be OR'd with preset numbers below to indicate F9 is initially ON
uint8_t Conf_PC_Num[125] = {
      0, 1, 2, 3, 4,
      5, 6, 7, 8, 9,
     10, 11, 12, 13, 14,
     15, 16, 17, 18, 19,
     20, 21, 22, 23, 24,
     25, 26, 27, 28, 29,
     30, 31, 32, 33, 34,
     35, 36, 37, 38, 39,
     40, 41, 42, 43, 44,
     45, 46, 47, 48, 49,
     50, 51, 52, 53, 54,
     55, 56, 57, 58, 59,
     60, 61, 62, 63, 64,
     65, 66, 67, 68, 69,
     70, 71, 72, 73, 74,
     75, 76, 77, 78, 79,
     80, 81, 82, 83, 84,
     85, 86, 87, 88, 89,
     90, 91, 92, 93, 94,
     95, 96, 97, 98, 99,
    100, 101, 102, 103, 104,
    105, 106, 107, 108, 109,
    110, 111, 112, 113, 114,
    115, 116, 117, 118, 119,
    120, 121, 122, 123, 124}; //<5> to <129> inclusive

// initial FX states for each PC (25 banks each of 5 patches) - 8 bits (low to high) for F1 to F8 (excluding Tap)
// F9 initial state is stored with the patch number as above
// note that even momentary FX fsw can be on by default - in which case, pressing sends their 'off' value(s), releasing sends their 'on' value(s)
uint8_t Conf_Init_FX_States[125] = {
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000};    //<130> to <254> inclusive

// Relay states for banks A1, A2 , A3, ..., E4, E5 - combine RELAYn_ constants
// these states are set on patch selection, and before CC resets (which can also include relay state changes)
// see below for information on relay settings
uint8_t Conf_Relay_Bank[25] = {
    R1_OFF | R2_OFF, REL_ALL_OFF, R1_ON | R2_OFF, R1_OFF | R2_ON, R1_ON | R2_ON,
    REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF,
    REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF,
    REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF,
    REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF, REL_ALL_OFF};    //<255> to <279> inclusive

// selecting a patch (incl backtracking) sets FX shift OFF? Stored below as PatchFxLo in Conf_Flags_11 byte

// ****** EFFECT CCs ****** (aka "IAs")
// NOTE CC numbers should not be 0 or 32, use CC numbers 1 to 31, 33 to 127, 128 for relays, or 255 for none
// There are 9 effects (excluding Tap, which will be inserted into this list later)
// Conf_FX_CC[0][n] to Conf_FX_CC[8][n] are for F1 to F9
#define CC_TAP     9  // Conf_FX_CC[9][n] is for Tap press/release - see below for more info
#define CC_TAP_FX 10  // Conf_FX_CC[10][n] is for additional (optional) TapFX CCs - see below for more info
#define CC_HEEL1  11  // Conf_FX_CC[11][n] is for heel-down CCs on pedal #1 - see below for more info
#define CC_HEEL2  12  // Conf_FX_CC[12][n] is for heel-down CCs on pedal #2 - see below for more info
// Each FX fsw can send 2 x CC messages, each with its own CC number, off-value and on-value
// Data order below is: 1st CC number, off-value, on-value, 2nd CC number, off-value, on-value
uint8_t Conf_FX_CC[13][6] = {
    {25, 0, 127, NONE, 0, 0},
    {26, 0, 127, NONE, 0, 0},
    {36, 127, 0, NONE, 0, 0},
    {29, NONE, 0, 28, 0, 127},
    {18, 0, 127, NONE, 0, 0},
    {113, 0, 127, NONE, 0, 0},
    {27, 0, 127, NONE, 0, 0},
    {50, 0, 127, NONE, 0, 0},
    {17, 127, 0, NONE, 0, 0},
    {64, NONE, 127, 65, 0, 127},
    {29, NONE, 127, 28, 0, 127},
    {69, 127, 0, NONE, 0, 0},
    {NONE, 0, 0, NONE, 0, 0} };  //<280> to <357> inclusive
/*
   for info only, these are the FX switched:
      {25, 0, 127, 255, 0, 0}     // F1: Drive
      {26, 0, 127, 7, 255, 127}   // F2: Boost, & set max vol when turned on
      {36, 127, 0, 255, 0, 0}     // F3: Reverb X/Y and Pre-Volume
      {29, 255, 0, 28, 0, 127}    // F4: Delay X/Y & bypass  (CC#29 is Delay X/Y)
      {18, 0, 127, 255, 0, 0}     // F5: Pitch & Volume FX
      {113, 0, 127, 255, 0, 0}    // F6: Tremolo
      {27, 0, 127, 255, 0, 0}     // F7: Phaser & Flanger
      {50, 0, 127, 255, 0, 0}     // F8: Chorus & Rotary/EQ
      {17, 127, 0, 255, 0, 0}     // F9: Alt-FX / Main-FX (eg Flange instead of Phase, Rotary instead of Chorus, etc)

      {64, NONE, 127, 65, 0, 127} // Tap press/release
      {29, NONE, 127, 28, 0, 127} // Additional Tap CCs after 2 presses
      {69, 127, 0, NONE, 0, 0}    // Heel-down CCs for pedal #1
      {NONE, 0, 0, NONE, 0, 0}    // Heel-down CCs for pedal #2
   note: my wah and similar effects are set in the Axe-FX to turn on automatically when pedal #2 is moved
   (pedal has a spring return to heel-down position)
*/

/*  RELAY SETTINGS
    All CC messages can also control relays as well. For example, this double CC message could be for an effect footswitch:
    {36, 0, 127, RELAYS, R2_OFF | R4_ON, R1_ON | R2_ON | R4_OFF}
    There are 2 messages here, firstly CC number 36 has an off value of 0 and an on value of 127
    Then there is a message for RELAYS, with an off value of "R2_OFF | R4_ON" and an on value of "R1_ON | R2_ON | R4_OFF"
    The "or" symbol "|" is used to combine on and off settings for relays 1 to 4
    So here's what this RELAYS example does:
    Turning the effect ON will
    - turn relay 1 on
    - turn relay 2 on
    - leave relay 3 unchanged, because it is not mentioned
    - turn relay 4 off
    Turning the effect OFF will
    - leave relay 1 unchanged, because it's not mentioned (in this case it would remain on from when this effect was turned on)
    - turn relay 2 off
    - leave relay 3 unchanged, because it's not mentioned
    - turn relay 4 on
    When a relay is ON, it is connected to the sleeve connection of the relay socket
*/

// which FX fsw are momentary? (excluding Tap, which is always momentary)
// F1 (bit0) to F8 (bit7) stored here, F9 is stored below as F9Mom in Conf_Flags_11 byte
uint8_t Conf_FX_Momentary = 0b00000000;  //<358>

// which FX CCs will be sent after PC based on initial or backtracked states? (excluding Tap)
// F1 (bit0) to F8 (bit7) stored here, F9 is stored below as F9AfterPC in Conf_Flags_11 byte
uint8_t Conf_Send_FX_After_PC = 0b00001111; //<359>

// which FX fsw are persistent? - these retain their states between patch changes (excluding backtracking)
// Their state will be sent when a NEW patch is selected and the effect state is non-default (or forced anyway)
// Persistence only applies to toggled effects
// F1 (bit0) to F8 (bit7) stored here, F9 is stored below as F9Pers in Conf_Flags_11 byte
uint8_t Conf_FX_Persistent = 0b00000000;  //<360>

// ****** TAP SETTINGS ******
// which is the Tap FX button (1 to 10)?  It is inserted into the list 9 of effects above
uint8_t Conf_Tap_FX_Num = 5;   //<361> must be 1 to 10

// Tap button 2 x CC messages - on for press, off for release are specified above in Conf_FX_CC

// Tap is a momentary footswitch (usually configured to send a CC message on press only)
// In addition, this MFC can send extra TapFX CC messages after a certain number of Tap presses after a patch change
// TapFX CC messages are specified above in Conf_FX_CC
// (TapFX OFF messages are sent if TapFX is turned off by another effect in the same ME group)
// This can be used to turn on an effect automatically after tap tempo has been set (also turns on the Tap LED)
uint8_t Conf_Tap_Count = 2;  //<362> 0 for no extra Tap messages, or 1 to 9 to send additional messages after that many tap presses (after a patch change)

// ****** MUTUALLY EXCLUSIVE EFFECTS ******
/*  Mutually exclusive (ME) effects work so that only one effect is on at the same time
    For example, if you include phaser and flange effects in a ME group and phaser is on, turning on the flanger will automatically turn off the phaser
    This MFC has 4 ME groups which you could use, for example, for modulation effects, drive/boost effects, delay effects and AxeFX scenes
    You can even include effects in more than one ME group if you want to get creative

    ME groups don't control which effects are on when patches are selected (or backtracked)
    They only turn off effects when another effect is turned on in the same group
    Other effects are turned off before the selected effect is turned on

    ME groups apply to toggled effects (where you press a footswitch once to turn it on, press again to turn it off)
    This includes the Tap footswitch when using Tap FX (where you specify a number of Tap presses to turn on Tap FX)
*/

// nibbles below for F2 (msn) & F1 (lsn), F4 & F3, F6 & F5, F8 & F7, TapFX (msn) & F9 (lsn)
// the four groups are represented by bits w, x, y & z in each byte: 0bwxyzwxyz
uint8_t Conf_ME_Groups[5] = {0b00000000, 0b00010000, 0b00000000, 0b00100010, 0b00010000};  //<363> to <367> inclusive
// settings above have Tap and Delay (F4) is the same ME group, and Phaser (F7) and Chorus (F8) are in another group

// ****** PEDAL SETTINGS ******
// CC number for pedals #1 & #2 (must be 1 to 31, 33 to 127, or 255 for none)
uint8_t Conf_Pedal_CC_Num[2] = {7, 4};        //<368> to <369> inclusive

// CC values to send after program changes for each pedal (0 to 127 or 255 for none or 254 to send previous pedal value)
uint8_t Conf_Pedal_After_PC[2] = {NONE, 0};   //<370> to <371> inclusive

// Send CC for pedals when heel down?  eg turn on wah pedal, or turn on tuner when pedal is heel-down on volume pedal
// these are specified above in Conf_FX_CC
// heel-down sends OFF values, not heel-down sends ON values (as always, swap on and off values for opposite switching)

// A dead area can be defined for pedal extremes, to ensure max and min values are always sent, and allow room for under-pedal switches
uint8_t Conf_Pedal_Toe_Dead_Pct[2] = {5, 5};   //<372> to <373> inclusive - pedal #1 and #2 % dead area of toe excursion (0 to 25)
uint8_t Conf_Pedal_Heel_Dead_Pct[2] = {10, 5}; //<374> to <375> inclusive - pedal #1 and #2 % dead area of heel excursion (0 to 25)

// The following are used to scale ADC data to CC range 0 to 127 - set from config calibration (default is entire byte range for ADC)
#define EEP_PEDAL_CALIB  2044    // start of EEPROM memory address for these settings (saved at end of EEPROM memory)
uint8_t Conf_Pedal_ADC_Min[2] = {0, 0};        //<2044> to <2045> inclusive
uint8_t Conf_Pedal_ADC_Max[2] = {255, 255};    //<2047> to <2048> inclusive

// For pedal faults, the following pedal values can be sent when pedals are disabled with the "pedal panic" option
// If a value is set here (ie not NONE) *and* the pedal has a heel CC message, it will also be sent (OFF messages if value here = 0, else sends ON messages)
uint8_t Conf_Pedal_Panic[2] = {127, 0};  //<376> to <377> inclusive - values for pedals 1 & 2 using pedal CC numbers above

// ****** BACK-TRACKING OPTIONS ******
// The [BACK] fsw toggles between the current and previous patch, optionally including effect states
// Backtracking enabled? If false, [BACK] goes immediately to BANK mode - Stored below as BackEnabled in Conf_Flags_11 byte

// Backtracking also recalls previous patch's effect states? Stored below as BackFx in Conf_Flags_11 byte

// Backtracking also selects the bank containing the back-tracked patch? Stored below as BackBank in Conf_Flags_11 byte

// ****** CUSTOM MESSAGE ******
// prior to a patch selection, the bottom row 6 fsw can select some CC messages to be sent after each patch change
// this can be used, for example, to set an ambient reverb level (or for Axe-FX, select a patch scene)
// values below set reverb levels 0%, 30%, 40%, 60%, 70%, 80% for CC number 16
// two messages can be sent (messages below send different values on the same CC number to force a change)
uint8_t Conf_Custom_CC_Num[2] = {16, 16};   //<378> to <379> CC number (or can use RELAYS) for 1st and 2nd custom messages
// then CC values (or relay settings) for first and second CC messages
// if a custom message is DE-selected, the *first* dual CC messages are sent below
// the next dual CC message values match their footswitch number 1 to 6
uint8_t Conf_Custom_CC_Val[7][2] = {{56, 64}, {8, 0}, {30, 38}, {43, 51}, {68, 76}, {81, 89}, {94, 102}};  //<380> to <393> inclusive

// ****** FLAGS ******
// see above for use of these flags - combine flags required in var Conf_Flags_11 (flags for settings version 1.1)
#define PatchFxLo   0b00000001  // to set FX as unshifted after patch change
#define F9AfterPC   0b00000010  // to always send F9 state after patch change
#define F9Mom       0b00000100  // if F9 is momentary
#define F9Pers      0b00001000  // if F9 is persistent
#define BackEnabled 0b00010000  // to enable patch backtracking
#define BackFX      0b00100000  // backtrack effect states as well
#define BackBank    0b01000000  // backtrack to bank containing backtracked patch
#define CustomLast  0b10000000  // to send custom CC messages after CC resets (after a patch change), else, custom messages sent before CC resets
uint8_t Conf_Flags_11 = PatchFxLo | F9AfterPC | BackEnabled | BackFX | CustomLast;  //<394>

// ****** HARDWARE ******
// How many msec to wait after sending a MIDI PC message, before sending CC messages (0 to 255)
uint8_t Conf_Pause_After_PC = 10; //<395>

// ****** AXE-FX II ******
// lowest 4 bits are used for Axe-FX model number (use one only - see axefx.h)
#define AxeFxTuner  0b10000000  // to show AxeFX tuner when MIDI in msg received
#define AxeFxTempo  0b01000000  // for backtracking to set previous patch global tempo
uint8_t Conf_AxeFX = AXEFX_2_MODEL | AxeFxTuner | AxeFxTempo;  //<396>

// ===================
// VARIABLES
// ===================

// vars starting with a lower case b hold boolean true1/false0 values
uint8_t PatchNow;        // current preset - this is the MFC preset number 0 to 124 (not the MIDI PC number)
uint8_t PatchPrev;       // previous preset, used for backtracking (with F9 state in highest bit) - start with invalid 127 for "no prev preset"
uint16_t FxNow;          // bits 0 to 9 track current FX states (with Tap inserted into FX order)
#define HEEL_FX1 10      // bits 10 & 11 of FxNow track pedal heel effect states for pedals #1 and #2
#define HEEL_FX2 11
uint16_t FxPrev;         // bits 0 to 9 track previous patch's FX states, used for backtracking (with Tap inserted into FX order)
uint8_t BankNow;         // currently selected bank 0 to 24 (not necessarily the same bank containing PathNow)
uint8_t BankGroup;       // used in BANK mode to track which group (A to E) is selected
ModeType ModeNow;        // set to one of the mode constants
ModeType ModePrev;       // set to one of the mode constants, to return from BANK mode to previous PLAY mode or STARTUP mode
uint8_t bFxShift;        // =true when FX shift activated in PLAY mode
uint8_t CustomMsg;       // custom message 1 to 6, or 0 for none
uint8_t TapCount;        // Tap press counter
uint8_t bWaitForPress;   // used to prevent a press in one mode activating a release in another mode
uint8_t LastPedalADC[2]; // last ADC values read from pedals #1 & #2
uint8_t LastPedalCC[2][2];  // last 2 x CC values sent for pedals #1 and #2 (set invalid values to force first CC send with changed ADC value)
uint8_t bSetupDirty;     // true when setup changes have been saved and MFC must restart
ButtonEvent lastButtonEvent;
#define TUNER_START 3    // decremented at approx 7.5Hz, so this is about 1/4 to 3/8 second
uint8_t LastTunerX;      // last LCD location a tuning arrow was shown
uint8_t bUsedEEPROM;     // true if settings were read from EEPROM, false for default settings

// Axe-FX tempo SysEx
uint8_t SysExTempo[16] = {0xF0, 0x00, 0x01, 0x74, 0x03, 0x02, 0x0D, 0x01, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0xF7};
                            // F0 start of sysex
                            // 00 01 74 manufacturer ID (Fractal Audio)
#define OFFSET_MODEL 4      // 03 is model number (AxeFX II)
#define OFFSET_FUNCTION 5   // 02 is function ID to get/set parameter
#define OFFSET_EFFECT_LSB 6 // 0D - effect ID LSB (7 bits: 0 to 6)
#define OFFSET_EFFECT_MSB 7 // 01 - effect ID MSB (high 7 bits) - 141 is "controllers"
#define OFFSET_PARAM_LSB 8  // 20 - parameter ID LSB
#define OFFSET_PARAM_MSB 9  // 00 - parameter ID MSB - 0020 (=32) is tempo parameter for AxeFX II
#define OFFSET_VALUE_LSB 10 // 1E to 7A - value LSB  (11th byte)
#define OFFSET_VALUE_MSB 11 // 00 to 01 - value MSB  (12th byte) - valid data values are 30 to 250
                            // 00 - extra value high 2 bits if needed
#define OFFSET_GET_SET 13   // 00 to request or 01 to set
                            // (Axe-FX reply inserts additional text information here)
#define OFFSET_CHECKSUM 14  // checksum for get/set messages
                            // F7 end of sysex
// AxeFxTempoBytes[x][y]  x=0 for prev tempo, x=1 for current tempo, y=0 for tempo lsb, y=1 for tempo msb
uint8_t AxeFxTempoBytes[2][2] = {{120, 0}, {120, 0}};
// countdown timer for replies
#define TEMPO_START 2    // decremented at approx 7.5Hz, so this is about 1/8 to 1/4 second

// vars modified by ISR
uint8_t volatile ChaseCount;      // counter for chasing LEDs
uint8_t volatile TunerCountdown;  // when tuner message received, countdown from value TUNER_START then resume normal LCD display
uint8_t volatile TempoCountdown;  // when tempo requested, wait briefly for a reply

// constant array vars
const uint8_t PedalPort[2] = {3, 4};   // pedal ports for pedals #1 and #2
const uint16_t Cylon[5] = {0b1100000000000001, 0b0111000000000111, 0b0001110000011100, 0b0000011101110000, 0b0000000111000000};
// tuner notes A to A-flat
const uint8_t TunerNote[12] = {'A', 'B', 'B', 'C', 'D', 'D', 'E', 'E', 'F', 'G', 'G', 'A'}; // A, Bb, B, C, Db, D, Eb, E, F, Gb, G, Ab
// 63 is in-tune, display in-tune over this range:
#define IN_TUNE_MIN 62
#define IN_TUNE_MAX 64
// assuming values are +/- 50 cents, then data range is approx 13 to 113
// divisions for showing the tuning progresssively more flat or sharp (more sensitive near to in-tune than out of tune):
const uint8_t TunerRange[2][6] = {{58, 53, 47, 40, 32, 23}, {68, 73, 79, 86, 94, 103}};

// values derived from config settings
uint8_t FswToFx[10];     // index from buttons, bFxNow bits etc to config arrays for F1 to F9 (values 0 to 8) and Tap (value 9)
uint8_t FxToFsw[10];     // index from F1[0] ... F9[8] then Tap[9] to button order, bFxNow bits etc
uint8_t Pedal_Active_Min[2];
uint8_t Pedal_Active_Max[2];
uint8_t Pedal_Active_Range[2];
uint8_t bPedalEnabled[2]; // true if not configured as NONE and if not manually (& temporarily) disabled with pedal panic
uint8_t bTempoBacktracking;

// ===================
// FUNCTIONS
// ===================

uint8_t GetBitInValue(uint8_t var, uint8_t bitnum)
{
  return (var >> bitnum) & 1;
}

uint8_t GetBitInValue16(uint16_t var, uint8_t bitnum)
{
  return (var >> bitnum) & 1;
}

uint16_t SetBitInValue16(uint16_t var, uint8_t newstate, uint8_t bitnum)
{
  if(newstate)
    return var | (1 << bitnum);
  else
    return var & ~(1 << bitnum);
}

void SetRelayStates(uint8_t states)
{
  uint8_t i;
  for (i = 0; i < 4; ++i) {
    if((states >> (i + 4)) & 1) {  // set relay state
      SetRelay(i, (states >> i) & 1);
    }
  }
}

void LcdClearLine(uint8_t row)
{
  LCDWriteCharsXY(0, row, ' ', 16);
}

void LcdShowBank()
{
  LCDDataXY(6, 0, (BankNow / 5) + 65);  // show bank group A to E
  LCDData((BankNow % 5) + 49);          // number 1 to 5
}

void LcdShowShift()
{
  LCDGotoXY(10, 0);
  if(bFxShift && !TunerCountdown)
    LCDData('\1');   // '\1' = right arrow
  else
    LCDData(' ');
}

void LcdShowPatchFX(uint8_t ShowPatch)
{
  uint8_t i;
  uint8_t g;
  uint8_t row;

  // show patch?
  if(ShowPatch && !TunerCountdown) { // eg A1-3 means bank A1, patch 3
    LCDDataXY(6, 1, (PatchNow / 25) + 65);  // show bank group A to E
    LCDData(((PatchNow / 5) % 5) + 49);     // bank number 1 to 5
    LCDData('-');
    LCDData((PatchNow % 5) + 49);           // patch number 1 to 5
  }

  // always show FX unless tuner is displayed
  if(!TunerCountdown) {
    LcdShowShift();
    for (g = 0; g < 10; g += 5) {       // for unshifted then shifted FX
      row = 1 - (g / 5);
      for (i = 0; i < 5; ++i) {
        if(GetBitInValue16(FxNow, g + i)) {
          if(FswToFx[g + i] == TAP_INDEX)    // tap state
            LCDDataXY(11 + i, row, 'T');
          else
            LCDDataXY(11 + i, row, FswToFx[g + i] + 49);   // number 1 to 9
        }
        else {
          LCDDataXY(11 + i, row, '\6');   // '\6' = centered dot
        }
      }
    }
  }
}

void SetLED(uint8_t LedNum, uint8_t bIsOn, LedColor OnColour)
{
  if(bIsOn)
    ledSetColor(LedNum, OnColour, false);
  else
    ledSetColor(LedNum, COLOR_BLACK, false);
}

void ShowLEDs(uint8_t LedGroup)
{ // set LEDs in current mode

  uint8_t i;
  uint8_t FxOffset;

  switch(ModeNow) {

  case MODE_STARTUP:

    if(LedGroup & LEDS_1_TO_5) {
      for (i = 0; i < 5; ++i)  // should also use with LEDS_6_AND_12 because all 6 fsw on bottom row can set custom msg
        SetLED(i, (CustomMsg == i + 1), COLOR_YELLOW);
    }

    if(LedGroup & LEDS_7_TO_11) {
      // all LEDs off (no patch selected yet)
      for (i = 6; i < 11; ++i)
        ledSetColor(i, COLOR_BLACK, false);
    }

    if(LedGroup & LEDS_6_AND_12) {
      ledSetColor(11, COLOR_BLACK, false);        //Mode LED is off
      SetLED(5, (CustomMsg == 6), COLOR_YELLOW);  //CustomMsg is 1-based
    }
    break;

  case MODE_BANK:

    if(LedGroup & LEDS_1_TO_5) {
      // show bank group A to E
      for (i = 0; i < 5; ++i)
        SetLED(i, (BankGroup == i), COLOR_YELLOW);
    }

    if(LedGroup & LEDS_7_TO_11) {
      // show bank number 1 to 5
      for (i = 0; i < 5; ++i)
        SetLED(i + 6, (BankNow % 5 == i), COLOR_YELLOW);
    }

    if(LedGroup & LEDS_6_AND_12) {
      // ledSetColor(11, COLOR_RED, false); // Mode LED will be flashed by Timer1 interrupt
      ledSetColor(5, COLOR_BLACK, false);   // off
    }
    break;

  default:  // PLAY mode

    FxOffset = 0;
    if(LedGroup & LEDS_1_TO_5) {
      if(bFxShift)
        FxOffset = 5;
      for (i = 0; i < 5; ++i)
        SetLED(i, (GetBitInValue16(FxNow, i + FxOffset)), COLOR_GREEN);
    }

    if(LedGroup & LEDS_7_TO_11) {
      for (i = 0; i < 5; ++i)
        SetLED(i + 6, (PatchNow % 5 == i), COLOR_RED);
    }

    if(LedGroup & LEDS_6_AND_12) {
      ledSetColor(11, COLOR_BLACK, false0);  //Mode LED is off
      ledSetColor(5, COLOR_BLACK, false0);   //FX Shift LED will be flashed by Timer1 interrupt if FX are shifted
    }
    break;

  }
  ledSend();
}

void ModeSet(ModeType NewMode)
{
  // NOTE this should only be called to set startup, play and bank modes!
  if(NewMode == MODE_BANK && ModeNow != MODE_BANK) {
    ModePrev = ModeNow;
    BankGroup = BankNow / 5;
  }

  //mode changes all occur on a press, so don't do anything in a new mode until another press occurs
  bWaitForPress = (ModeNow == NewMode) ? false0 : true1;

  // set current mode
  ModeNow = NewMode;

  // show lcd screen for new mode
  if(TunerCountdown)
    LcdClearLine(0);
  else
    LCDClear();

  LCDSetCursor(LS_NONE);
  LCDWriteTextXY(1, 0, "Bank");
  LcdShowBank();

  switch(ModeNow) {

  case MODE_STARTUP:
    if(TunerCountdown) {
      LCDWriteTextXY(9, 0, "STARTUP");
    }
    else {
      LCDWriteTextXY(0, 1, Display_Version);
      if(bUsedEEPROM)
        LCDDataXY(13, 1, '\6');  // show centered dot in version number if EEPROM was used
    }
    break;

  case MODE_BANK:
    if(TunerCountdown) {
      LCDWriteTextXY(12, 0, "BANK");
    }
    else {
      LCDDataXY(11, 0, '\2'); //up arrow
      LCDWriteText("1-5");
      LCDWriteTextXY(2, 1, "Select");
      LCDDataXY(11, 1, '\3'); //down arrow
      LCDWriteText("A-E");
    }
    break;

  default:  // PLAY mode
    if(TunerCountdown) {
      LCDWriteTextXY(12, 0, "PLAY");
    }
    else {
      LCDWriteTextXY(0, 1, "Patch");
      LcdShowPatchFX(true1);
    }
  }

  // show LEDs in new mode
  ShowLEDs(LEDS_ALL);
}

uint8_t CC_Valid_Data(uint8_t val, uint8_t bIsCCNum, uint8_t bIsCCNumB)
{
  // validates configured CC numbers (1-31, 33-127 or 255) and values (0-127 or 255)
  if(val == 128 && bIsCCNumB)
    return val; // ok - 2nd CC number 128 is interpreted as a relay "message"
  else if(val & 128)
    return 255; // return 255 (none) if val >= 128
  else if(!bIsCCNum)
    return val; // CC values 0 to 127 are ok
  else if(val == 0 || val == 32)
    return 255; // CC numbers cannot be 0 or 32
  else
    return val;
}

void InitVars()
{
  uint8_t f;
  uint8_t i;
  uint16_t wTemp;

  // =======================================
  // initialise global vars
  // =======================================

  PatchNow = 0;
  PatchPrev = 0;
  FxNow = 0;
  FxPrev = 0;
  BankNow = 0;
  BankGroup = 0;
  ModeNow = MODE_STARTUP;
  ModePrev = MODE_STARTUP;
  bFxShift = false0;
  CustomMsg = 0;
  TapCount = 0;
  ChaseCount = 0;
  bWaitForPress = false0;
  bSetupDirty = false0;
  LastPedalCC[0][0] = 255;
  LastPedalCC[0][1] = 255;
  LastPedalCC[1][0] = 255;
  LastPedalCC[1][1] = 255;
  TunerCountdown = 0;
  TempoCountdown = 0;
  bUsedEEPROM = false0;

  for (i = 0; i < 2; ++i) {
    for (f = 0; f < 2; ++f)
      AxeFxTempoBytes[f][i] = 0;
  }

  // init tempo sysex
  SysExTempo[OFFSET_MODEL] = Conf_AxeFX & 0x0F;     // AxeFX II is 0x03
  //TODO - not sure if other models are the same here
  // SysExTempo[OFFSET_EFFECT_LSB] = effect_ID_LSB;   // AxeFX II is 0x0D
  // SysExTempo[OFFSET_EFFECT_MSB] = effect_ID_MSB;   // AxeFX II is 0x01
  // SysExTempo[OFFSET_PARAM_LSB] = parameter_ID_LSB; // AxeFX II is 0x0D
  // SysExTempo[OFFSET_PARAM_MSB] = parameter_ID_MSB; // AxeFX II is 0x01

  // =====================================================
  // read settings from EEPROM if correct settings version
  // =====================================================

  f = eeprom_read_byte((uint8_t *)0);
  i = eeprom_read_byte((uint8_t *)1);
  if(f == Conf_Settings_Version[0] && i == Conf_Settings_Version[1]) {
    bUsedEEPROM = true1;
    // read settings
    Conf_MIDI_Channel = eeprom_read_byte((uint8_t *)2);
    Conf_MIDI_Bank_MSB = eeprom_read_byte((uint8_t *)3);
    Conf_MIDI_Bank_LSB = eeprom_read_byte((uint8_t *)4);

    for (i = 0; i < 124; ++i)
      Conf_PC_Num[i] = eeprom_read_byte((uint8_t *)5 + i);

    for (i = 0; i < 124; ++i)
      Conf_Init_FX_States[i] = eeprom_read_byte((uint8_t *)130 + i); // 130 to 254

    for (i = 0; i < 25; ++i)
      Conf_Relay_Bank[i] = eeprom_read_byte((uint8_t *)255 + i);

    for (i = 0; i < 13; ++i) {
      for (f = 0; f < 6; ++f)
        Conf_FX_CC[i][f] = eeprom_read_byte((uint8_t *)280 + (i * 6) + f);
    }

    Conf_FX_Momentary = eeprom_read_byte((uint8_t *)358);
    Conf_Send_FX_After_PC = eeprom_read_byte((uint8_t *)359);
    Conf_FX_Persistent = eeprom_read_byte((uint8_t *)360);
    Conf_Tap_FX_Num = eeprom_read_byte((uint8_t *)361);
    Conf_Tap_Count = eeprom_read_byte((uint8_t *)362);

    for (i = 0; i < 5; ++i)
      Conf_ME_Groups[i] = eeprom_read_byte((uint8_t *)363 + i);

    for (i = 0; i < 2; ++i) { //WARNING: this data is not read in address order
      Conf_Pedal_CC_Num[i] = eeprom_read_byte((uint8_t *)368 + i);
      Conf_Pedal_After_PC[i] = eeprom_read_byte((uint8_t *)370 + i);
      Conf_Pedal_Toe_Dead_Pct[i] = eeprom_read_byte((uint8_t *)372 + i);
      Conf_Pedal_Heel_Dead_Pct[i] = eeprom_read_byte((uint8_t *)374 + i);
      Conf_Pedal_Panic[i] = eeprom_read_byte((uint8_t *)376 + i);
    }

    for (i = 0; i < 2; ++i) { //WARNING: this data is not read in address order
      Conf_Custom_CC_Num[i] = eeprom_read_byte((uint8_t *)378 + i);
      for (f = 0; f < 7; ++f)
        Conf_Custom_CC_Val[f][i] = eeprom_read_byte((uint8_t *)380 + (f * 2) + i);
    }

    Conf_Flags_11 = eeprom_read_byte((uint8_t *)394);
    Conf_Pause_After_PC = eeprom_read_byte((uint8_t *)395);
    Conf_AxeFX = eeprom_read_byte((uint8_t *)396);
  }

  // read pedal calibration values from EEPROM
  Conf_Pedal_ADC_Min[0] = eeprom_read_byte((uint8_t *)EEP_PEDAL_CALIB);
  Conf_Pedal_ADC_Min[1] = eeprom_read_byte((uint8_t *)(EEP_PEDAL_CALIB + 1));
  Conf_Pedal_ADC_Max[0] = eeprom_read_byte((uint8_t *)(EEP_PEDAL_CALIB + 2));
  Conf_Pedal_ADC_Max[1] = eeprom_read_byte((uint8_t *)(EEP_PEDAL_CALIB + 3));

  // =======================================
  // validate config settings
  // =======================================

  Conf_MIDI_Channel &= 0x0F;
  Conf_MIDI_Bank_MSB = CC_Valid_Data(Conf_MIDI_Bank_MSB, false0, false0);
  Conf_MIDI_Bank_LSB = CC_Valid_Data(Conf_MIDI_Bank_LSB, false0, false0);

  for (f = 0; f < 13; ++f) {
    for (i = 0; i < 6; ++i) {
      if(i == 0)
        Conf_FX_CC[f][i] = CC_Valid_Data(Conf_FX_CC[f][i], true1, false0);
      else if(i == 3)
        Conf_FX_CC[f][i] = CC_Valid_Data(Conf_FX_CC[f][i], true1, true1);
      else if(Conf_FX_CC[f][(i / 3) * 3] < 128) // check CC values if "CC number" is not RELAYS or NONE
        Conf_FX_CC[f][i] = CC_Valid_Data(Conf_FX_CC[f][i], false0, false0);
    }
  }

  if(Conf_Tap_FX_Num < 1  || Conf_Tap_FX_Num > 10)
    Conf_Tap_FX_Num = 5;  // default

  if(Conf_Tap_Count > 9)
    Conf_Tap_Count = 0;   // 0 for none

  for (i = 0; i < 2; ++i) {
    // pedal CC number
    Conf_Pedal_CC_Num[i] = CC_Valid_Data(Conf_Pedal_CC_Num[i], true1, false0);  // pedal CC number cannot for 128 (for relay control)
    // pedal CC value to send after PC (254 is OK for "previous" value)
    if(Conf_Pedal_After_PC[i] != 254)
      Conf_Pedal_After_PC[i] = CC_Valid_Data(Conf_Pedal_After_PC[i], false0, false0);
    // pedal dead % ranges
    if(Conf_Pedal_Toe_Dead_Pct[i] > 25)
      Conf_Pedal_Toe_Dead_Pct[i] = 25;
    if(Conf_Pedal_Heel_Dead_Pct[i] > 25)
      Conf_Pedal_Heel_Dead_Pct[i] = 25;
    // panic pedal CC value
    Conf_Pedal_Panic[i] = CC_Valid_Data(Conf_Pedal_Panic[i], false0, false0);
    // pedal ADC min and max (min must be <= max)
    if(Conf_Pedal_ADC_Min[i] >= Conf_Pedal_ADC_Max[i]) {
      Conf_Pedal_ADC_Min[i] = 0;
      Conf_Pedal_ADC_Max[i] = 255;
    }
  }

  for (i = 0; i < 2; ++i) {
    Conf_Custom_CC_Num[i] = CC_Valid_Data(Conf_Custom_CC_Num[i], true1, (i == 0) ? false0 : i);  // allow relays for 2nd custom msg
    if(Conf_Custom_CC_Num[i] != 128) {  // check CC value if not for relays
      for (f = 0; f < 7; ++f)
        Conf_Custom_CC_Val[f][i] = CC_Valid_Data(Conf_Custom_CC_Val[f][i], false0, false0);
    }
  }

  // =======================================
  // now set vars derived from config values
  // =======================================

  // load index arrays for FswToFx[] and FxToFsw[]
  f = 0;
  for (i = 0; i < 10; ++i) {
    if(i == Conf_Tap_FX_Num - 1) {
      f = 1;		// =1 when found Tap effect
      FswToFx[i] = 9;
      FxToFsw[9] = i;
    }
    else {
      FswToFx[i] = i - f;  // 0 to 8 for F1 to F9
      if(i < 9)
        FxToFsw[i] = i + f;
    }
  }

  // pedal vars
  for (i = 0; i < 2; ++i) {
    // pre-calculate vars used to convert pedal ADC values to CC values
    wTemp = ((Conf_Pedal_ADC_Max[i] - Conf_Pedal_ADC_Min[i] + 1) * Conf_Pedal_Heel_Dead_Pct[i]) / 100;  // dead ADC range for pedal heel
    Pedal_Active_Min[i] = Conf_Pedal_ADC_Min[i] + wTemp;
    wTemp = ((Conf_Pedal_ADC_Max[i] - Conf_Pedal_ADC_Min[i] + 1) * Conf_Pedal_Toe_Dead_Pct[i]) / 100;  // dead ADC range for pedal toe
    Pedal_Active_Max[i] = Conf_Pedal_ADC_Max[i] - wTemp;
    Pedal_Active_Range[i] = Pedal_Active_Max[i] - Pedal_Active_Min[i] + 1;
    // also check pedal enable, and read initial ADC value
    bPedalEnabled[i] = ~Conf_Pedal_CC_Num[i] ? true1 : false0;  // true if Conf_Pedal_CC_Num[i] != NONE where NONE=255 meaning no CC
    LastPedalADC[i] = adcRead8MsbBit(PedalPort[i]);  // take a first measurement to stabilise
  }

  // tempo backtracking
  bTempoBacktracking = (Conf_Flags_11 & BackEnabled) && (Conf_AxeFX & AxeFxTempo) ? true1 : false0;

  // set startup mode
  ModeSet(MODE_STARTUP);
}

void SendTempoSysEx()
{
  // set checksum
  uint8_t i;
  uint8_t tmp = SysExTempo[0];
  for (i = 1; i < OFFSET_CHECKSUM; ++i)
    tmp ^= SysExTempo[i];
  SysExTempo[OFFSET_CHECKSUM] = tmp & 0x7F;
  // send sysex message
  for (i = 0; i < 16; ++i)
    uart0PutChar(SysExTempo[i]);
}

void AxeFxRequestTempo() {
  // tempo value LSB & MSB are ignored for request
  SysExTempo[OFFSET_GET_SET] = 0;  // request value
  TempoCountdown = TEMPO_START;  // start countdown
  SendTempoSysEx();
}

void AxeFxSetTempo() {
  // sets current tempo to AxeFxTempoBytes[1][y] after backtracking has swapped prev and current values
  if(AxeFxTempoBytes[1][0] || AxeFxTempoBytes[1][1]) {  // don't set tempo if lsb & msb bytes == 0
    SysExTempo[OFFSET_VALUE_LSB] = AxeFxTempoBytes[1][0];  // tempo LSB
    SysExTempo[OFFSET_VALUE_MSB] = AxeFxTempoBytes[1][1];  // tempo MSB
    SysExTempo[OFFSET_GET_SET] = 1;   // set value
    SendTempoSysEx();
  }
}

void sysExCallback(uint16_t length)
{ // callback for incoming sysex messages from axefx

  uint8_t i;
  uint8_t* sysexData;
  uint8_t NewTunerX;

  //check if SysEx from Axe Fx, if not do nothing
  sysexData = midiGetLastSysExData();    //get pointer to last sysex message payload data
  if(!axeFxCheckFractalManufId(sysexData))
    return;

  AxeFxEffectTunerInfo tunerMsg;
  if(axeFxGetFunctionId(sysexData) == AXEFX_TUNER_INFO && (Conf_AxeFX & AxeFxTuner)) {
    if(ModeNow == MODE_STARTUP || ModeNow == MODE_PLAY || ModeNow == MODE_BANK) { // not a setup mpode
      if(!TunerCountdown) {  // switch to tuner display
        TunerCountdown = TUNER_START;  // set TunerCountdown != 0
        ModeSet(ModeNow); // this displays top LCD line with MODE if TunerCountdown != 0
        LCDWriteCharsXY(0, 1, '\6', 16);  // '\6' for centered dots shown on tuner row
        LastTunerX = 16;  // an invalid location to flag no last arrow and prevent NewTunerX == LastTunerX
      }
      // and always reset countdown when new message received
      TunerCountdown = TUNER_START;
      // display tuning data
      axefxParseTunerInfo(&tunerMsg, sysexData);
      // show tuner note
      tunerMsg.note_ = tunerMsg.note_ % 12;   // % 12 prevents out of array bounds
      LCDDataXY(7, 1, TunerNote[tunerMsg.note_]);
      // show flat or string number
      LCDGotoXY(8, 1);
      switch(tunerMsg.note_)
      {
      case 1:
      case 4:
      case 6:
      case 9:
      case 11:
        LCDData('\7');  // '\7' for flat symbol
        break;
      default:
        LCDData(tunerMsg.stringNumber_ + 49);   // show string number if note is not a flat
      }

      // calc tuning arrow position
      if(tunerMsg.tuneData_ < IN_TUNE_MIN) {  // flat
        for (i = 0; i < 6; ++i)
          if(tunerMsg.tuneData_ > TunerRange[0][i]) break;
        NewTunerX = 6 - i;
      }
      else if(tunerMsg.tuneData_ > IN_TUNE_MAX) {  //sharp
        for (i = 0; i < 6; ++i)
          if(tunerMsg.tuneData_ < TunerRange[1][i]) break;
        NewTunerX = 9 + i;
      }
      else {  // in tune
        NewTunerX = 8;  // =8 for "in tune"
      }

      // erase last arrow, show new arrow
      if(LastTunerX != NewTunerX) {
        // erase last arrow(s)
        if(LastTunerX == 8) {
          LCDDataXY(6, 1, '\6');  // '\6' is centered dot
          LCDDataXY(9, 1, '\6');
        }
        else if(LastTunerX != 16) {
          LCDDataXY(LastTunerX, 1, '\6');  // centered dot
        }
        // show new arrow(s)
        if(NewTunerX == 8) {
          LCDDataXY(6, 1, '\1');  // '\1' is right arrow
          LCDDataXY(9, 1, '\0');  // '\0' is left arrow
        }
        else if(NewTunerX < 8) {
          LCDDataXY(NewTunerX, 1, '\1');  // '\1' is right arrow
        }
        else {
          LCDDataXY(NewTunerX, 1, '\0');  // '\0' is left arrow
        }
        LastTunerX = NewTunerX;
      }
    }
  }

  else if(bTempoBacktracking && TempoCountdown) { // sysex msg received within timeout
    //check received sysex length
    if(midiGetSysExLength(sysexData) >= 16) {
      // check model number, function id, effect id and parameter ids are for tempo
      for (i = OFFSET_MODEL; i <= OFFSET_PARAM_MSB; ++i){
        if(*(sysexData + i) != SysExTempo[i])
          break;
      }
      if(i > OFFSET_PARAM_MSB) { // didn't exit above for loop early
        // capture current tempo msb & lsb
        AxeFxTempoBytes[1][0] = *(sysexData + OFFSET_VALUE_LSB);  // current tempo LSB
        AxeFxTempoBytes[1][1] = *(sysexData + OFFSET_VALUE_MSB);  // current tempo MSB
        TempoCountdown = 0; // clear countdown
      }
    }
  }
}

void SendCC(uint8_t CC_Num, uint8_t CC_Val)
{
  if(CC_Num == 128)
    SetRelayStates(CC_Val);
  else if(CC_Num != 255 && CC_Val != 255)
    midiSendControlChange(CC_Num, CC_Val, Conf_MIDI_Channel);
}

void SendFxCC(uint8_t FxIndex, uint8_t IsOn)
{
  SendCC(Conf_FX_CC[FxIndex][0], Conf_FX_CC[FxIndex][1 + IsOn]);
  SendCC(Conf_FX_CC[FxIndex][3], Conf_FX_CC[FxIndex][4 + IsOn]);
}

void SendCustCC()
{
  SendCC(Conf_Custom_CC_Num[0], Conf_Custom_CC_Val[CustomMsg][0]);
  SendCC(Conf_Custom_CC_Num[1], Conf_Custom_CC_Val[CustomMsg][1]);
}

void SendPatch()
{
  // PatchNow and FxNow are already set

  uint8_t bBackTap = false0;   // true if tap FX messages need to be sent after other FX CC states messages
  uint8_t i;

  // send MIDI bank MSB & LSB?
  SendCC(0, Conf_MIDI_Bank_MSB);
  SendCC(32, Conf_MIDI_Bank_LSB);

  // send patch
  midiSendProgramChange(Conf_PC_Num[PatchNow], Conf_MIDI_Channel);  //this function ignores bit 7 of PC number

  // short pause after PC message?
  i = Conf_Pause_After_PC;
  while(i--) {
    _delay_ms(1); // must use a loop for variable delay, because _delay_ms() expects a constant
  }

  // send custom msg?
  if(CustomMsg && !(Conf_Flags_11 & CustomLast))
    SendCustCC();

  // set bank relay states
  SetRelayStates(Conf_Relay_Bank[PatchNow / 5]); // relay config for bank of current patch

  // send F1 to F9 FX states after PC (if forced or non-default)
  for (i = 0; i < 10; ++i)
  {
    if(FswToFx[i] == TAP_INDEX) {
      // send TapFX messages if Tap is on (back-tracked)
      if(Conf_Tap_Count && GetBitInValue16(FxNow, i))
        bBackTap = true1;
    }
    else if(FswToFx[i] == FX9_INDEX) {
      // send F9 FX state after PC (if forced or non-default) - default state is in bit 7 of MIDI program number
      if((Conf_Flags_11 & F9AfterPC) || GetBitInValue16(FxNow, i) != GetBitInValue(Conf_PC_Num[PatchNow], 7))
        SendFxCC(FX9_INDEX, GetBitInValue16(FxNow, i));
    }
    else if(GetBitInValue(Conf_Send_FX_After_PC, FswToFx[i]) || GetBitInValue16(FxNow, i) != GetBitInValue(Conf_Init_FX_States[PatchNow], FswToFx[i]))
      SendFxCC(FswToFx[i], GetBitInValue16(FxNow, i)); // send F1 to F8 state after PC (if forced or non-default)
  }

  // send pedal messages after PC?
  for (i = 0; i < 2; ++i) {
    if(bPedalEnabled[i]) {
      if(Conf_Pedal_After_PC[i] == PREV)
        SendCC(Conf_Pedal_CC_Num[i], LastPedalCC[0][i]);
      else
        SendCC(Conf_Pedal_CC_Num[i], Conf_Pedal_After_PC[i]);
    }
  }

  // send tap FX CCs (Tap can be on when backtracking)
  if(bBackTap) {
    SendFxCC(CC_TAP_FX, 1); // already checked that TapFX is enabled
    TapCount = Conf_Tap_Count;
  }
  else {
    TapCount = 0;
  }

  // send custom msg?
  if(CustomMsg && (Conf_Flags_11 & CustomLast))
    SendCustCC();

  // unshift FX after patch selection?
  if(Conf_Flags_11 & PatchFxLo)
    bFxShift = false0;

  // reset countdowns
  TunerCountdown = 0;
  TempoCountdown = 0;

  if(ModeNow == MODE_PLAY)   // (LCD and LEDs will be updated from startup mode when mode is changed to play mode)
    ModeSet(ModeNow);
}

uint8_t IsMomentary(uint8_t FxInd)
{
  // FxInd = 0 to 8 for F1 to F9, =9 for Tap
  if(FxInd == TAP_INDEX)  // this is the Tap fsw
    return (Conf_Tap_Count) ? false0 : true1; // if Tap FX not enabled, Tap is treated as a momentary fsw
  else if(FxInd == FX9_INDEX)
    return (Conf_Flags_11 & F9Mom) ? true1 : false0;
  else
    return GetBitInValue(Conf_FX_Momentary, FxInd);
}

uint8_t FxInGroup(uint8_t FxInd, uint8_t GroupInd)
{
  // FxInd = 0 to 8 for F1 to F9, =9 for Tap
  uint8_t me_flags = Conf_ME_Groups[FxInd / 2];
  if((FxInd & 1) == 1)
    me_flags >>= 4;  // group flags are in msn
  return ((me_flags >> GroupInd) & 1) ? true1 : false0;
}

void ResetTapCount()
{
  // if TapFX is enabled but not yet activated, reset Tap counter
  // Tap must be pressed a preset number of times without different button presses in-between
  if(Conf_Tap_Count && !GetBitInValue16(FxNow, FxToFsw[TAP_INDEX]))
    TapCount = 0;
}

uint8_t NoOffMsg(uint8_t FxInd) {
  // FxInd is 0 to 8 for F1 to F9
  // leave F1 to F9 on for toggles with no off messages - assists ME grouped toggles that have only on messages (eg AxeFX scenes)
  // returns true if FxNum has no off message (for both messages, either no message or CC message with no off value)
  if (Conf_FX_CC[FxInd][0] == NONE || (Conf_FX_CC[FxInd][0] < 128 && Conf_FX_CC[FxInd][1] == NONE))
    return (Conf_FX_CC[FxInd][3] == NONE || (Conf_FX_CC[FxInd][3] < 128 && Conf_FX_CC[FxInd][4] == NONE)) ? true : false;
  else
    return false0;
}

void SendFx(uint8_t FxNum, uint8_t bFswPress)
{
  // FxNum is 0 to 9 in actual effect order (with Tap inserted)
  uint8_t FxInd = FswToFx[FxNum];   // =0 to 8 to FxNum is F1 to F9, =9 for FxNum is Tap
  uint8_t bIsMom = IsMomentary(FxInd);  // true if FxInd effect is momentary (or Tap with no TapFX configured)
  uint8_t NewState;
  uint8_t i;
  uint8_t f;

  // ===============================
  // CHECK MUTUALLY EXCLUSIVE GROUPS
  // ===============================

  // if this is a fsw press about to turn a toggle (or Tap FX) effect ON,
  // we need to find out if other effect(s) in the same mutually exclusive group need to be turned OFF first
  if (bFswPress && !bIsMom) {
    if(FxInd == TAP_INDEX)  // this is the Tap FX fsw, configured to turn on after configured presses
      NewState = ((TapCount + 1) == Conf_Tap_Count) ? true1 : false0;
    else if(NoOffMsg(FswToFx[FxNum]))  // already tested not momentary
      NewState = 1; // toggle fsw with no off values always send on msg with press (and turns on if not already)
    else
      NewState = 1 - GetBitInValue16(FxNow, FxNum); // will toggle state

    if(NewState) {                        // about to turn on TapFX or toggle FX - check each ME group
      for (i = 0; i < 4; ++i) {           // for each ME group
        if(FxInGroup(FxInd, i)) {         // FxInd is in group i
          for (f = 0; f < 10; ++f) {      // for F1 to F9 then Tap
            if(f != FxInd) {              // not the current effect being turned on
              if(FxInGroup(f, i)) {       // effect f is also in group i, find location for f
                if(GetBitInValue16(FxNow, FxToFsw[f])) {   // effect is on
                  if(f == TAP_INDEX) {
                    if(Conf_Tap_Count) {    // turn it off if Tap FX enabled
                      FxNow = SetBitInValue16(FxNow, 0, FxToFsw[f]);
                      TapCount = 0;
                      SendFxCC(CC_TAP_FX, 0);  // send tap off fx
                    }
                  }
                  else if(!IsMomentary(f)) {  // not momentary and not Tap fsw - so turn off toggle effect
                    FxNow = SetBitInValue16(FxNow, 0, FxToFsw[f]);
                    SendFxCC(f, 0);
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // ===================
  // APPLY FX CHANGES
  // ===================

  if(FxInd == TAP_INDEX) {
    // for tap, always send on value(s) for press, off value(s) for release
    SendFxCC(CC_TAP, bFswPress);

    if(Conf_Tap_Count) {
      if(bFswPress && (TapCount < Conf_Tap_Count)){
        ++TapCount;
        if(TapCount == Conf_Tap_Count) {
          // configured presses after a PC turns on tap and sends additional CC messages
          FxNow = SetBitInValue16(FxNow, 1, FxNum);
          SendFxCC(CC_TAP_FX, 1);
        }
      }
    }
    else {
      // tap FX not enabled, works like a momentary fsw, off by default, on when pressed
      FxNow = SetBitInValue16(FxNow, bFswPress, FxNum);
    }
    // get Axe-FX tempo?
    if (bTempoBacktracking && bFswPress)
      AxeFxRequestTempo();
  }
  else {
    // action fsw change?
    ResetTapCount();
    if(bFswPress || bIsMom) {
      // NOTE: condition above is (bFswPress || bIsMom) and condition below is (!bIsMom && NoOffMsg(...))
      // therefore condition below is actually: (bFswPress && !bIsMom && NoOffMsg(...))
      //   always send on msg for toggle presses with no off msg
      //   else toggle state for all presses and for release of momentary fsw
      NewState = (!bIsMom && NoOffMsg(FswToFx[FxNum])) ? 1 : 1 - GetBitInValue16(FxNow, FxNum);
      FxNow = SetBitInValue16(FxNow, NewState, FxNum);
      SendFxCC(FxInd, NewState);
    }
  }

  ShowLEDs(LEDS_1_TO_5);
  LcdShowPatchFX(false0); // update FX only on LCD
}

void ButtonEventStartup(ButtonEvent button)
{
  uint8_t i;
  // only fsw press events apply in STARTUP mode
  if(button.actionType_ == BUTTON_PUSH) {
    bWaitForPress = false0;
    if(button.buttonNum_ < 6) {
      // prior to preset selection, fsw 1 to 6 set/unset custom message on press event
      // custom messages are previewed here in STARTUP mode, and sent after each patch change in PLAY mode
      if((Conf_Custom_CC_Num[0] < 128 && Conf_Custom_CC_Val[button.buttonNum_ + 1][0] != NONE)
          || (Conf_Custom_CC_Num[1] < 128 && Conf_Custom_CC_Val[button.buttonNum_ + 1][1] != NONE)
          || Conf_Custom_CC_Num[1] == RELAYS) {
        // one or both custom messages are set
        if(CustomMsg == button.buttonNum_ + 1)  // already set to this number, so turn it off and send "off-preview" message
          CustomMsg = 0;
        else     // set this index and send preview message
          CustomMsg = button.buttonNum_ + 1;
        SendCustCC();
        ShowLEDs(LEDS_1_TO_5 | LEDS_6_AND_12);  // update LEDS 1 to 6 in STARTUP mode
      }
    }
    else if(button.buttonNum_ == 11) {
      ModeSet(MODE_BANK);   // go to BANK mode
    }
    else if(button.buttonNum_ < 11) {
      // patch change sent with fsw 7 to 11 (buttonEvent.buttonNum_ = 6 to 10)
      // this is the first patch, so use default fx states and set prev patch and prev FX to same for backtracking
      PatchNow = (BankNow * 5) + button.buttonNum_ - 6;
      PatchPrev = PatchNow;
      for (i = 0; i < 10; ++i) {
        if(FswToFx[i] == TAP_INDEX) {  // Tap
          FxNow = SetBitInValue16(FxNow, 0, i);
          TapCount = 0;     // clear tap counter
        }
        else if(FswToFx[i] == FX9_INDEX) {
          FxNow = SetBitInValue16(FxNow, GetBitInValue(Conf_PC_Num[PatchNow], 7), i);  // FX9 state is in highest bit of PC number
        }
        else {
          FxNow = SetBitInValue16(FxNow, GetBitInValue(Conf_Init_FX_States[PatchNow], FswToFx[i]), i);
        }
      }
      FxPrev = FxNow;
      SendPatch();
      ModeSet(MODE_PLAY);
    }
  }
}

void ButtonEventBank(ButtonEvent button)
{
  // only fsw press events apply in BANK mode
  if(button.actionType_ == BUTTON_PUSH) {
    bWaitForPress = false0;
    if(button.buttonNum_ < 5) {
      // bank group selected with fsw 1 to 5 (buttonEvent.buttonNum_ 0 to 4)
      BankGroup = button.buttonNum_;
      LCDDataXY(6, 0, BankGroup + 65);  // preview selected bank group A to E
      ShowLEDs(LEDS_1_TO_5);
    }
    else if(button.buttonNum_ == 5) {
      // select bank for current patch and return to previous mode
      if(ModePrev != MODE_STARTUP) {
        // a patch has been selected
        BankNow = PatchNow / 5;
        ModeSet(ModePrev);
      }
    }
    else if(button.buttonNum_ == 11) {
      ModeSet(ModePrev);  // return to previous mode with no bank change
    }
    else if(button.buttonNum_ < 11) {
      // bank number selected with fsw 7 to 11 (buttonEvent.buttonNum_ = 6 to 10)
      BankNow = (BankGroup * 5) + button.buttonNum_ - 6;
      ModeSet(ModePrev);
    }
  }
}

void ButtonEventPlay(ButtonEvent button)
{
  uint8_t FxNum = 0;
  uint8_t i;
  uint8_t iTmp;
  uint16_t bTmp;

  if(button.actionType_ == BUTTON_PUSH) {
    bWaitForPress = false0;
    if(button.buttonNum_ < 5) {
      // pressed FX 1 to 5
      if(bFxShift)
        FxNum = 5;
      FxNum += button.buttonNum_;
      SendFx(FxNum, true1); // also updates FX LEDs
    }
    else if(button.buttonNum_ == 5) {
      ResetTapCount();
      bFxShift = bFxShift ? false0 : true1;  //toggle FX shift
      LcdShowShift();
      ShowLEDs(LEDS_1_TO_5 | LEDS_6_AND_12);
    }
    else if(button.buttonNum_ == 11) {
      // go immediately to BANK mode if back-tracking is not enabled
      ResetTapCount();
      if(!(Conf_Flags_11 & BackEnabled))
        ModeSet(MODE_BANK);
    }
    else if(button.buttonNum_ < 11) {
      // patch change sent with fsw 7 to 11 (buttonEvent.buttonNum_ = 6 to 10)
      // this is a changed (or reset) patch, so use current patch & fx states for backtracking
      ResetTapCount();
      PatchPrev = PatchNow;
      PatchNow = (BankNow * 5) + button.buttonNum_ - 6;
      FxPrev = FxNow;
      for (i = 0; i < 10; ++i) {
        if(FswToFx[i] == TAP_INDEX) { // Tap
          FxNow = SetBitInValue16(FxNow, 0, i);
          TapCount = 0;   // reset Tap counter
        }
        else if(FswToFx[i] == FX9_INDEX) {
          // set default FX state unless persistent (when patch changes, configured persistent and not a momentary fsw)
          if(PatchNow == PatchPrev || !(Conf_Flags_11 & F9Pers) || IsMomentary(FX9_INDEX))
            FxNow = SetBitInValue16(FxNow, GetBitInValue(Conf_PC_Num[PatchNow], 7), i);  // FX9 state is in highest bit of PC number
        }
        else {
          // set default FX state unless persistent (when patch changes, configured persistent and not a momentary fsw)
          if(PatchNow == PatchPrev || !GetBitInValue(Conf_FX_Persistent, FswToFx[i]) || IsMomentary(FswToFx[i]))
            FxNow = SetBitInValue16(FxNow, GetBitInValue(Conf_Init_FX_States[PatchNow], FswToFx[i]), i);
        }
      }
      // keep tempo history & clear current tempo
      for (i = 0; i < 2; ++i) {
        AxeFxTempoBytes[0][i] = AxeFxTempoBytes[1][i];
        AxeFxTempoBytes[1][i] = 0;
      }
      SendPatch();
    }
  }
  else if(button.actionType_ == BUTTON_RELEASE && !bWaitForPress) {
    //fsw released (only used for releasing momentary FX fsw, including Tap)
    if(button.buttonNum_ < 5) {
      if(bFxShift)
        FxNum = 5;
      FxNum += button.buttonNum_;
      SendFx(FxNum, false0);
    }

    else if((button.buttonNum_ == 11) && (Conf_Flags_11 & BackEnabled)) {
      // if backtrack is enabled, and [BACK] was released before being held long enough to select BANK mode
      // swap current patch for previous patch
      iTmp = PatchPrev;
      PatchPrev = PatchNow;
      PatchNow = iTmp;

      // swap FX now for previous if enabled, else use default FX states
      bTmp = FxPrev;  // by default, unless not using backtracked FX
      for (i = 0; i < 10; ++i) {

        if(FswToFx[i] == TAP_INDEX)
          TapCount = 0;   // reset Tap counter

        if(!(Conf_Flags_11 & BackFX)) {  // not using backtracked FX
          if(FswToFx[i] == FX9_INDEX)
            bTmp = SetBitInValue16(bTmp, GetBitInValue(Conf_PC_Num[PatchNow], 7), i);
          else
            bTmp = SetBitInValue16(bTmp, GetBitInValue(Conf_Init_FX_States[PatchNow], FswToFx[i]), i);
        }
      }
      FxPrev = FxNow;
      FxNow = bTmp | (FxNow & ((1 << HEEL_FX1) | (1 << HEEL_FX2)));  // preserve heel fx states

      if(bTempoBacktracking) {
        // swap current and prev tempos (LSB and MSB)
        for (i = 0; i < 2; ++i) {
          iTmp = AxeFxTempoBytes[0][i];
          AxeFxTempoBytes[0][i] = AxeFxTempoBytes[1][i];
          AxeFxTempoBytes[1][i] = iTmp;
        }
      }

      // set bank for previous patch if enabled
      if(Conf_Flags_11 & BackBank) {
        BankNow = PatchNow / 5;
        LcdShowBank();
      }

      // and send the new patch
      SendPatch();

      // backtrack tempo?
      if(bTempoBacktracking)
        AxeFxSetTempo();
    }
  }
  else if(button.actionType_ == BUTTON_HOLDON && !bWaitForPress)
  {
    if(button.buttonNum_ == 11 && (Conf_Flags_11 & BackEnabled))
      ModeSet(MODE_BANK);   // go to BANK mode if back-tracking is enabled and [BACK] held down
  }
}

void WaitForRelease()
{
  do
  {
    lastButtonEvent = getButtonLastEvent();
  } while(lastButtonEvent.actionType_ != BUTTON_RELEASE);
}

void LcdShowPanicAction(uint8_t action)
{
  LCDGotoXY(5, 1);
  switch(action) {
  case 0:
    LCDWriteText("NONE");
    break;
  case 1:
    LCDWriteText("PED1");
    break;
  case 2:
    LCDWriteText("PED2");
    break;
  default:
    LCDWriteText("BOTH");
  }
  LCDGotoXY(5, 1);
}

void PedalPanic()
{
  uint8_t PanicAction = 3;  // 0=none, 1=pedal1, 2=pedal2, 3=both
  uint8_t bSetup = true1;
  uint8_t i;
  uint8_t HeelState;

  // setup mode (but this is only a temporary config change, not saved)
  ModePrev = ModeNow;
  ModeNow = MODE_SETUP;
  TunerCountdown = 0;
  TempoCountdown = 0;
  ledSetColorAll(COLOR_BLACK, true);    // all LEDs off

  // enter pedal panic screen
  LCDClear();
  LCDWriteTextXY(0, 0, "Disable Pedals");
  LCDWriteTextXY(10, 1, "OK/NO");
  LcdShowPanicAction(PanicAction);
  LCDSetCursor(LS_ULINE);

  // pedal disable loop
  while(bSetup)
  {
    lastButtonEvent = getButtonLastEvent();
    if(lastButtonEvent.actionType_ == BUTTON_PUSH) {

      switch(lastButtonEvent.buttonNum_) {

      case KEY_UP:
        PanicAction = (PanicAction + 3) % 4;
        LcdShowPanicAction(PanicAction);
        break;

      case KEY_DOWN:
        PanicAction = (PanicAction + 1) % 4;
        LcdShowPanicAction(PanicAction);
        break;

      case KEY_LOAD_OK:
        for (i = 0; i < 2; ++i) {
          if(!(PanicAction & (i + 1))) {
            bPedalEnabled[i] = ~Conf_Pedal_CC_Num[i] ? true1 : false0;  // true if Conf_Pedal_CC_Num[i] != NONE where NONE=255 meaning no CC
          }
          else {
            bPedalEnabled[i] = false0;
            if((Conf_Pedal_CC_Num[i] != NONE) && (Conf_Pedal_Panic[i] != NONE)) {
              HeelState = (Conf_Pedal_Panic[i]) ? true1 : false0;
              FxNow = SetBitInValue16(FxNow, HeelState, HEEL_FX1 + i);  // set heel fx state
              SendFxCC(CC_HEEL1 + i, HeelState);  // send heel state CCs
              SendCC(Conf_Pedal_CC_Num[i], Conf_Pedal_Panic[i]);   // send pedal value
            }
          }
        }
        // no break, continue below
      case KEY_SETUP_NO:
        bSetup = false0;
        break;
      }
    }
  }

  ModeSet(ModePrev);
  WaitForRelease();
}

void PedalCalibrate()
{
  uint8_t PedalNum = 0;
  uint8_t TempMin[2];
  uint8_t TempMax[2];
  uint8_t PedalADC;
  uint8_t i;
  uint8_t bSetup = true1;

  // init temp vars in case only one pedal (or no pedals) set before [OK] saves
  for (i = 0; i < 2; ++i) {
    TempMin[i] = Conf_Pedal_ADC_Min[i];
    TempMax[i] = Conf_Pedal_ADC_Max[i];
  }

  LCDClear();
  LCDWriteTextXY(0, 0, "Pedal1 Calibrate");
  LCDDataXY(10, 1, '\1');  // '\1' is right arrow
  LCDWriteText("start");
  LCDGotoXY(5, 0);
  LCDSetCursor(LS_ULINE);

  // SETUP LOOP
  while(bSetup)
  {
    // action a keypress
    lastButtonEvent = getButtonLastEvent();
    if(lastButtonEvent.actionType_ == BUTTON_PUSH) {

      switch(lastButtonEvent.buttonNum_) {

      case KEY_UP:
      case KEY_DOWN:
        if(ModeNow == MODE_SETUP){
          PedalNum = 1 - PedalNum;
          LCDDataXY(5, 0, PedalNum + 49);
          LCDGotoXY(5, 0);
        }
        break;

      case KEY_RIGHT:
        // start calibration mode, read & display first value
        if(ModeNow == MODE_SETUP) {
          ModeNow = MODE_SETUP_NOC;
          LCDSetCursor(LS_NONE);
          // briefly show current min and max
          LcdClearLine(1);
          LCDWriteIntXY(2, 1, Conf_Pedal_ADC_Min[PedalNum], 3);
          LCDWriteIntXY(12, 1, Conf_Pedal_ADC_Max[PedalNum], 3);
          _delay_ms(1000);
          // start with pedal current value
          PedalADC = adcRead8MsbBit(PedalNum + 3);   // pedal ports are 3 and 4
          TempMin[PedalNum] = PedalADC;
          TempMax[PedalNum] = PedalADC;
          LCDWriteIntXY(2, 1, PedalADC, 3);
          LCDWriteIntXY(7, 1, PedalADC, 3);
          LCDWriteIntXY(12, 1, PedalADC, 3);
          LCDDataXY(0, 1, '\0');    // '\0' is left arrow
        }
        break;

      case KEY_LEFT:
        // exit calibration mode
        if(ModeNow == MODE_SETUP_NOC){
          LcdClearLine(1);
          LCDDataXY(10, 1, '\1');  // '\1' is right arrow
          LCDWriteText("start");
          LCDGotoXY(5, 0);
          ModeNow = MODE_SETUP;
          LCDSetCursor(LS_ULINE);
        }
        break;

      case KEY_SETUP_NO:
        if(ModeNow == MODE_SETUP)
          bSetup = false0;
        break;

      case KEY_LOAD_OK:
        if(ModeNow == MODE_SETUP) {
          // save settings in local vars & update EEPROM
          for (i = 0; i < 2; ++i) {
            Conf_Pedal_ADC_Min[i] = TempMin[i];
            eeprom_update_byte((uint8_t *)(EEP_PEDAL_CALIB + i), Conf_Pedal_ADC_Min[i]);
            Conf_Pedal_ADC_Max[i] = TempMax[i];
            eeprom_update_byte((uint8_t *)(EEP_PEDAL_CALIB + 2 + i), Conf_Pedal_ADC_Max[i]);
          }
          // and exit loop
          bSetupDirty = true1;
          bSetup = false0;
        }
        //break;
      }
    }

    // display pedal values
    if(ModeNow == MODE_SETUP_NOC) {
      PedalADC = adcRead8MsbBit(PedalNum + 3);   // pedal ports are 3 and 4
      LCDWriteIntXY(7, 1, PedalADC, 3);
      if(PedalADC < TempMin[PedalNum]) {
        TempMin[PedalNum] = PedalADC;
        LCDWriteIntXY(2, 1, PedalADC, 3);
      }
      else if(PedalADC > TempMax[PedalNum]) {
        TempMax[PedalNum] = PedalADC;
        LCDWriteIntXY(12, 1, PedalADC, 3);
      }
    }
  }
}

void SetupMFC()
{
  // init setup
  bSetupDirty = false0;
  ModePrev = ModeNow;     // to go back to prev mode if no changes saved
  ModeNow = MODE_SETUP;
  TunerCountdown = 0;
  TempoCountdown = 0;
  ledSetColorAll(COLOR_BLACK, true);    // all LEDs off

  // calibrate pedals
  PedalCalibrate();

  // exit setup
  if(bSetupDirty) {
    InitVars();   // validate config settings, recalc dependent global vars, and set STARTUP mode
  }
  else {
    ModeSet(ModePrev);
  }
  WaitForRelease();
}

void InitTimer1()
{  // init timer 1 for LED flashing & chasing
  cli();                                    // Disable interrupts while initialising timer
  TCCR1B |= (1 << WGM12 );                  // Configure timer 1 for CTC mode
  TIMSK |= (1 << OCIE1A );                  // Enable CTC interrupt
  OCR1A = 1000;                             // = (FCPU / Prescale / required Hz) - 1 : here for 8Mhz, 1024 prescale, approx 7.5Hz
  TCCR1B |= ((1 << CS10) | (1 << CS12));    // prescale 1024
  TCCR1B &= ~(1 << CS11);
  sei();                                    // Enable global interrupts
}

ISR(TIMER1_COMPA_vect)
{
  uint8_t i;

  switch(ModeNow) {

  case MODE_PLAY:
    i = (BankNow == PatchNow / 5);
    if(bFxShift) {  // flash LED 6 red in PLAY mode
      if(ChaseCount & 1)
        ledSetColor(5, COLOR_RED, i);
      else
        ledSetColor(5, COLOR_BLACK, i);
    }
    if(!i) {  // flash patch LED if not in current bank
      if(ChaseCount & 1)
        ledSetColor(6 + (PatchNow % 5), COLOR_BLACK, true);
      else
        ledSetColor(6 + (PatchNow % 5), COLOR_RED, true);
    }
   break;

  case MODE_BANK:
    SetLED(11, (ChaseCount & 1), COLOR_RED);
    ledSend();
    break;

  case MODE_STARTUP:
    // chase patch LEDs "cylon style"
    for (i = 0; i < 5; ++i)
      SetLED(i + 6, GetBitInValue16(Cylon[i], ChaseCount), COLOR_RED);
    ledSend();
    break;

  case MODE_SETUP:
    if(ChaseCount & 3)
      LCDSetCursor(LS_ULINE);
    else
      LCDSetCursor(LS_NONE);
    break;

  case MODE_SETUP_NOC:
    break;

  }
  ChaseCount = (ChaseCount + 1) & 0x0F; //0-15 counter

  // revert from tuner display?
  if(TunerCountdown) {
    --TunerCountdown;
    if(!TunerCountdown)
      ModeSet(ModeNow);  //resume normal display
  }

  // tempo countdown
  if(TempoCountdown)
    --TempoCountdown;
}

int main(void)
{
  uint8_t i;

  // pedal tracking
  uint8_t PedalADC;
  uint8_t PedalCC;

  // Library initialization
  initBjDevLib();
  InitRelays();      // Set relay outputs
  LCDInit(LS_NONE);  // LCD library initialisation

  // initialisations
  InitVars();         // Init MFC vars and sets STARTUP mode
  InitTimer1();       // Init timer 1

  // register midi callback for SysEx messages
  midiRegisterSysExCallback(sysExCallback);

  // MAIN LOOP
  while(1)
  {
    // buttons
    lastButtonEvent = getButtonLastEvent();
    if(lastButtonEvent.actionType_ != BUTTON_NO_EVENT) {
      if(lastButtonEvent.buttonNum_ == KEY_LOAD_OK) {
        if(lastButtonEvent.actionType_ == BUTTON_PUSH)
          PedalPanic();
      }
      else if(lastButtonEvent.buttonNum_ == KEY_SETUP_NO) {
        if(lastButtonEvent.actionType_ == BUTTON_PUSH)
          SetupMFC();
      }
      else if(ModeNow == MODE_PLAY) {
        ButtonEventPlay(lastButtonEvent);
      }
      else if(ModeNow == MODE_STARTUP) {
        ButtonEventStartup(lastButtonEvent);
      }
      else if(ModeNow == MODE_BANK) {
        ButtonEventBank(lastButtonEvent);
      }
    }

    // expression pedals
    for (i = 0; i < 2; ++i) {

      if(bPedalEnabled[i]) {

        PedalADC = adcRead8MsbBit(PedalPort[i]);
        if(PedalADC != LastPedalADC[i]) {
          LastPedalADC[i] = PedalADC;

          // calculate CC value
          if(PedalADC <= Pedal_Active_Min[i])
            PedalCC = 0;
          else if(PedalADC >= Pedal_Active_Max[i])
            PedalCC = 127;
          else  // this is a 16-bit calculation, but not necessary to force it
            PedalCC = ((PedalADC - Pedal_Active_Min[i]) << 7) / Pedal_Active_Range[i];   // <<7 is *128

          if(PedalCC != LastPedalCC[0][i]) {  // not same as last value sent
            // always send 0 and 127 values, or if not same as previous-to-last value
            if(PedalCC == 127 || PedalCC == 0 || PedalCC != LastPedalCC[1][i])
              midiSendControlChange(Conf_Pedal_CC_Num[i], PedalCC, Conf_MIDI_Channel);
            // keep value history
            LastPedalCC[1][i] = LastPedalCC[0][i];
            LastPedalCC[0][i] = PedalCC;
            // check heel-down effect
            if(PedalCC) {
              if(!GetBitInValue16(FxNow, HEEL_FX1 + i)) {
                SendFxCC(CC_HEEL1 + i, 1);  // turn heel effect on
                FxNow = SetBitInValue16(FxNow, true1, HEEL_FX1 + i);
              }
            }
            else { // PedalCC == 0
              if(GetBitInValue16(FxNow, HEEL_FX1 + i)) {
                SendFxCC(CC_HEEL1 + i, 0);  // turn heel effect off
                FxNow = SetBitInValue16(FxNow, false0, HEEL_FX1 + i);
              }
            }
          }  // end if(PedalCC != LastPedalCC[i]) ...
        }  // end if(PedalADC != LastPedalADC[i] ...
      }  // end if(bPedalEnabled[i]) ...
    }  // end for (i = 0; i < 2; ++i) ...

    // check MIDI input
    midiRead();  // this actually returns bool true/false

  }  // end while(1) ...
}